@@ -10,14 +10,18 @@ import time
 import inspect
 import argparse
 import traceback
+from copy import copy
+import binwalk
 import binwalk.core.statuserver
 import binwalk.core.common
 import binwalk.core.settings
 import binwalk.core.plugin
-from threading import Thread
 from binwalk.core.compat import *
+from binwalk.core.exceptions import *
+
 
 class Option(object):
+
     '''
     A container class that allows modules to declare command line options.
     '''
@@ -57,7 +61,8 @@ class Option(object):
 
     def convert(self, value, default_value):
         if self.type and (self.type.__name__ == self.dtype):
-            # Be sure to specify a base of 0 for int() so that the base is auto-detected
+            # Be sure to specify a base of 0 for int() so that the base is
+            # auto-detected
             if self.type == int:
                 t = self.type(value, 0)
             else:
@@ -69,7 +74,9 @@ class Option(object):
 
         return t
 
+
 class Kwarg(object):
+
     '''
     A container class allowing modules to specify their expected __init__ kwarg(s).
     '''
@@ -88,7 +95,9 @@ class Kwarg(object):
         self.default = default
         self.description = description
 
+
 class Dependency(object):
+
     '''
     A container class for declaring module dependencies.
     '''
@@ -99,7 +108,9 @@ class Dependency(object):
         self.kwargs = kwargs
         self.module = None
 
+
 class Result(object):
+
     '''
     Generic class for storing and accessing scan results.
     '''
@@ -136,7 +147,9 @@ class Result(object):
         for (k, v) in iterator(kwargs):
             setattr(self, k, v)
 
+
 class Error(Result):
+
     '''
     A subclass of binwalk.core.module.Result.
     '''
@@ -152,7 +165,9 @@ class Error(Result):
         self.exception = None
         Result.__init__(self, **kwargs)
 
+
 class Module(object):
+
     '''
     All module classes must be subclassed from this.
     '''
@@ -168,13 +183,14 @@ class Module(object):
     # A list of default dependencies for all modules; do not override this unless you
     # understand the consequences of doing so.
     DEFAULT_DEPENDS = [
-            Dependency(name='General',
-                       attribute='config'),
-            Dependency(name='Extractor',
-                       attribute='extractor'),
+        Dependency(name='General',
+                   attribute='config'),
+        Dependency(name='Extractor',
+                   attribute='extractor'),
     ]
 
-    # A list of binwalk.core.module.Dependency instances that can be filled in as needed by each individual module.
+    # A list of binwalk.core.module.Dependency instances that can be filled in
+    # as needed by each individual module.
     DEPENDS = []
 
     # Format string for printing the header during a scan.
@@ -216,7 +232,8 @@ class Module(object):
     # Modules with a higher order are displayed first in help output
     ORDER = 5
 
-    # Set to False if this is not a primary module (e.g., General, Extractor modules)
+    # Set to False if this is not a primary module (e.g., General, Extractor
+    # modules)
     PRIMARY = True
 
     def __init__(self, parent, **kwargs):
@@ -318,6 +335,13 @@ class Module(object):
     def _plugins_pre_scan(self):
         self.plugins.pre_scan_callbacks(self)
 
+    def _plugins_load_file(self, fp):
+        try:
+            self.plugins.load_file_callbacks(fp)
+            return True
+        except IgnoreFileException:
+            return False
+
     def _plugins_new_file(self, fp):
         self.plugins.new_file_callbacks(fp)
 
@@ -351,7 +375,8 @@ class Module(object):
         # Calls the unload method for all dependency modules.
         # These modules cannot be unloaded immediately after being run, as
         # they must persist until the module that depends on them is finished.
-        # As such, this must be done separately from the Modules.run 'unload' call.
+        # As such, this must be done separately from the Modules.run 'unload'
+        # call.
         for dependency in self.dependencies:
             try:
                 getattr(self, dependency.attribute).unload()
@@ -375,7 +400,8 @@ class Module(object):
             except Exception:
                 pass
 
-        # Add any pending extracted files to the target_files list and reset the extractor's pending file list
+        # Add any pending extracted files to the target_files list and reset
+        # the extractor's pending file list
         self.target_file_list += self.extractor.pending
 
         # Reset all dependencies prior to continuing with another file.
@@ -389,7 +415,7 @@ class Module(object):
 
             # Values in self.target_file_list are either already open files (BlockFile instances), or paths
             # to files that need to be opened for scanning.
-            if isinstance(next_target_file, str):
+            if isinstance(next_target_file, str) or isinstance(next_target_file, unicode):
                 fp = self.config.open_file(next_target_file)
             else:
                 fp = next_target_file
@@ -397,7 +423,8 @@ class Module(object):
             if not fp:
                 break
             else:
-                if self.config.file_name_filter(fp) == False:
+                if (self.config.file_name_filter(fp) == False or
+                        self._plugins_load_file(fp) == False):
                     fp.close()
                     fp = None
                     continue
@@ -443,14 +470,15 @@ class Module(object):
         # Add the name of the current module to the result
         r.module = self.__class__.__name__
 
-        # Any module that is reporting results, valid or not, should be marked as enabled
+        # Any module that is reporting results, valid or not, should be marked
+        # as enabled
         if not self.enabled:
             self.enabled = True
-
         self.validate(r)
         self._plugins_result(r)
 
-        # Update the progress status automatically if it is not being done manually by the module
+        # Update the progress status automatically if it is not being done
+        # manually by the module
         if r.offset and r.file and self.AUTO_UPDATE_STATUS:
             self.status.total = r.file.length
             self.status.completed = r.offset
@@ -534,7 +562,8 @@ class Module(object):
         self.modules = self.parent.executed_modules
 
         # A special exception for the extractor module, which should be allowed to
-        # override the verbose setting, e.g., if --matryoshka has been specified
+        # override the verbose setting, e.g., if --matryoshka has been
+        # specified
         if hasattr(self, "extractor") and self.extractor.config.verbose:
             self.config.verbose = self.config.display.verbose = True
 
@@ -574,7 +603,9 @@ class Module(object):
 
         return retval
 
+
 class Status(object):
+
     '''
     Class used for tracking module status (e.g., % complete).
     '''
@@ -584,17 +615,12 @@ class Status(object):
         self.clear()
 
     def clear(self):
-        for (k,v) in iterator(self.kwargs):
+        for (k, v) in iterator(self.kwargs):
             setattr(self, k, v)
 
-class ModuleException(Exception):
-    '''
-    Module exception class.
-    Nothing special here except the name.
-    '''
-    pass
 
 class Modules(object):
+
     '''
     Main class used for running and managing modules.
     '''
@@ -628,19 +654,23 @@ class Modules(object):
     def __exit__(self, t, v, b):
         self.cleanup()
 
-    def _set_arguments(self, argv=[], kargv={}):
-        for (k,v) in iterator(kargv):
-            k = self._parse_api_opt(k)
-            if v not in [True, False, None]:
-                if not isinstance(v, list):
-                    v = [v]
-                for value in v:
-                    if not isinstance(value, str):
-                        value = str(bytes2str(value))
-                    argv.append(k)
-                    argv.append(value)
-            else:
-                argv.append(k)
+    def _set_arguments(self, argv=None, kargv=None):
+        if kargv:
+            for (k, v) in iterator(kargv):
+                    k = self._parse_api_opt(k)
+                    if v is not True and v is not False and v is not None:
+                        if not isinstance(v, list):
+                            v = [v]
+                        for value in v:
+                            if not isinstance(value, str):
+                                value = str(bytes2str(value))
+                            argv.append(k)
+                            argv.append(value)
+                    else:
+                        # Only append if the value is True; this allows for toggling values
+                        # by the function call.
+                        if v:
+                            argv.append(k)
 
         if not argv and not self.arguments:
             self.arguments = sys.argv[1:]
@@ -648,7 +678,8 @@ class Modules(object):
             self.arguments = argv
 
     def _parse_api_opt(self, opt):
-        # If the argument already starts with a hyphen, don't add hyphens in front of it
+        # If the argument already starts with a hyphen, don't add hyphens in
+        # front of it
         if opt.startswith('-'):
             return opt
         # Short options are only 1 character
@@ -672,6 +703,32 @@ class Modules(object):
             if inspect.isclass(module) and hasattr(module, attribute):
                 modules[module] = module.PRIORITY
 
+        # user-defined modules
+        try:
+            import imp
+            load_module_source = imp.load_source
+        except ImportError:
+            # Python 3.12 compatibility
+            import importlib.machinery
+            load_module_source = importlib.machinery.SourceFileLoader
+
+        user_modules = binwalk.core.settings.Settings().user.modules
+
+        for file_name in os.listdir(user_modules):
+            if not file_name.endswith('.py'):
+                continue
+            module_name = file_name[:-3]
+            try:
+                user_module = load_module_source(module_name, os.path.join(user_modules, file_name))
+            except KeyboardInterrupt as e:
+                raise e
+            except Exception as e:
+                binwalk.core.common.warning("Error loading module '%s': %s" % (file_name, str(e)))
+
+            for (name, module) in inspect.getmembers(user_module):
+                if inspect.isclass(module) and hasattr(module, attribute):
+                    modules[module] = module.PRIORITY
+
         return sorted(modules, key=modules.get, reverse=True)
 
     def help(self):
@@ -681,11 +738,16 @@ class Modules(object):
         Returns the help string.
         '''
         modules = {}
-        help_string = "\nBinwalk v%s\nCraig Heffner, http://www.binwalk.org\n" % binwalk.core.settings.Settings.VERSION
-        help_string += "\nUsage: binwalk [OPTIONS] [FILE1] [FILE2] [FILE3] ...\n"
+        help_string = "\n"
+        help_string += "Binwalk v%s\n" % binwalk.__version__
+        help_string += "Craig Heffner, ReFirmLabs\n"
+        help_string += "https://github.com/ReFirmLabs/binwalk\n"
+        help_string += "\n"
+        help_string += "Usage: binwalk [OPTIONS] [FILE1] [FILE2] [FILE3] ...\n"
 
         # Build a dictionary of modules and their ORDER attributes.
-        # This makes it easy to sort modules by their ORDER attribute for display.
+        # This makes it easy to sort modules by their ORDER attribute for
+        # display.
         for module in self.list(attribute="CLI"):
             if module.CLI:
                 modules[module] = module.ORDER
@@ -707,7 +769,7 @@ class Modules(object):
                     else:
                         short_opt = "   "
 
-                    fmt = "    %%s %%s%%-%ds%%s\n" % (25-len(long_opt))
+                    fmt = "    %%s %%s%%-%ds%%s\n" % (25 - len(long_opt))
                     help_string += fmt % (short_opt, long_opt, optargs, module_option.description)
 
         return help_string + "\n"
@@ -728,9 +790,11 @@ class Modules(object):
         for module in self.list():
             obj = self.run(module)
 
-        # Add all loaded modules that marked themselves as enabled to the run_modules list
+        # Add all loaded modules that marked themselves as enabled to the
+        # run_modules list
         for (module, obj) in iterator(self.executed_modules):
-            # Report the results if the module is enabled and if it is a primary module or if it reported any results/errors
+            # Report the results if the module is enabled and if it is a
+            # primary module or if it reported any results/errors
             if obj.enabled and (obj.PRIMARY or obj.results or obj.errors):
                 run_modules.append(obj)
 
@@ -750,16 +814,19 @@ class Modules(object):
                 self.status.clear()
 
             # If the module is not being loaded as a dependency, add it to the executed modules dictionary.
-            # This is used later in self.execute to determine which objects should be returned.
+            # This is used later in self.execute to determine which objects
+            # should be returned.
             if not dependency:
                 self.executed_modules[module] = obj
 
                 # The unload method tells the module that we're done with it, and gives it a chance to do
-                # any cleanup operations that may be necessary. We still retain the object instance in self.executed_modules.
+                # any cleanup operations that may be necessary. We still retain
+                # the object instance in self.executed_modules.
                 obj._unload_dependencies()
                 obj.unload()
         except KeyboardInterrupt as e:
-            # Tell the status server to shut down, and give it time to clean up.
+            # Tell the status server to shut down, and give it time to clean
+            # up.
             if self.status.running:
                 self.status.shutdown = True
                 while not self.status.finished:
@@ -778,9 +845,10 @@ class Modules(object):
         import binwalk.modules
         attributes = {}
 
-        for dependency in module.DEFAULT_DEPENDS+module.DEPENDS:
+        for dependency in module.DEFAULT_DEPENDS + module.DEPENDS:
 
-            # The dependency module must be imported by binwalk.modules.__init__.py
+            # The dependency module must be imported by
+            # binwalk.modules.__init__.py
             if hasattr(binwalk.modules, dependency.name):
                 dependency.module = getattr(binwalk.modules, dependency.name)
             else:
@@ -799,7 +867,8 @@ class Modules(object):
             if module_enabled or not dependency.kwargs:
                 depobj = self.run(dependency.module, dependency=True, kwargs=dependency.kwargs)
 
-            # If a dependency failed, consider this a non-recoverable error and raise an exception
+            # If a dependency failed, consider this a non-recoverable error and
+            # raise an exception
             if depobj.errors:
                 raise ModuleException("Failed to load " + dependency.name + " module")
             else:
@@ -816,19 +885,19 @@ class Modules(object):
 
         Returns a dictionary of kwargs for the specified module.
         '''
-        kwargs = {'enabled' : False}
+        kwargs = {'enabled': False}
         last_priority = {}
-        longs = []
-        shorts = ""
         parser = argparse.ArgumentParser(add_help=False)
         # Hack: This allows the ListActionParser class to correllate short options to long options.
-        #       There is probably a built-in way to do this in the argparse.ArgumentParser class?
+        # There is probably a built-in way to do this in the
+        # argparse.ArgumentParser class?
         parser.short_to_long = {}
 
         # Must build arguments from all modules so that:
         #
         #    1) Any conflicting arguments will raise an exception
-        #    2) The only unknown arguments will be the target files, making them easy to identify
+        #    2) The only unknown arguments will be the target files, making them
+        #       easy to identify
         for m in self.list(attribute="CLI"):
 
             for module_option in m.CLI:
@@ -848,7 +917,8 @@ class Modules(object):
                     parser_kwargs['action'] = 'store_true'
                 elif module_option.type is list:
                     parser_kwargs['action'] = 'append'
-                    parser.short_to_long[module_option.short] = module_option.long
+                    parser.short_to_long[
+                        module_option.short] = module_option.long
 
                 parser.add_argument(*parser_args, **parser_kwargs)
 
@@ -871,10 +941,12 @@ class Modules(object):
                 for (name, default_value) in iterator(module_option.kwargs):
 
                     # If this kwarg has not been previously processed, or if its priority is equal to or
-                    # greater than the previously processed kwarg's priority, then let's process it.
+                    # greater than the previously processed kwarg's priority,
+                    # then let's process it.
                     if not has_key(last_priority, name) or last_priority[name] <= module_option.priority:
 
-                        # Track the priority for future iterations that may process the same kwarg name
+                        # Track the priority for future iterations that may
+                        # process the same kwarg name
                         last_priority[name] = module_option.priority
 
                         try:
@@ -901,7 +973,7 @@ class Modules(object):
                 if has_key(kwargs, module_argument.name):
                     arg_value = kwargs[module_argument.name]
                 else:
-                    arg_value = module_argument.default
+                    arg_value = copy(module_argument.default)
 
                 setattr(obj, module_argument.name, arg_value)
 
@@ -927,6 +999,7 @@ class Modules(object):
             except Exception as e:
                 binwalk.core.common.warning("Failed to start status server on port %d: %s" % (port, str(e)))
 
+
 def process_kwargs(obj, kwargs):
     '''
     Convenience wrapper around binwalk.core.module.Modules.kwargs.
@@ -940,6 +1013,7 @@ def process_kwargs(obj, kwargs):
         kwargs = m.kwargs(obj, kwargs)
     return kwargs
 
+
 def show_help(fd=sys.stdout):
     '''
     Convenience wrapper around binwalk.core.module.Modules.help.
@@ -950,5 +1024,3 @@ def show_help(fd=sys.stdout):
     '''
     with Modules() as m:
         fd.write(m.help())
-
-
